from functools import wraps
from time import perf_counter as counter
from typing import Callable

from horology.tformatter import rescale_time


def timed(f: Callable = None, name=None, *, unit='auto', print_fn=print):
    """Decorator that prints time of execution of the decorated function

    Parameters
    ----------
    f: Callable
        The function which execution time should be measured.
    name: str or None, optional
        String that should be printed as the function name. By default
        the f.__name__ proceeded by a colon and space is used. See
        examples below.
    unit: {'auto', 'ns', 'us', 'ms', 's', 'min', 'h', 'd'}
        Time unit used to print elapsed time. Use 'a' or 'auto' for
        automatic time adjustment (default).
    print_fn: Callable or None, optional
        Function that is called to print the time elapsed. Use `None` to
        disable printing anything. You can provide e.g. `logger.info`.
        By default the built-in `print` function is used.

    Attributes
    ----------
    interval: float
        Time elapsed by the function in seconds. Can be used to get the
        time programmatically after the execution of f.

    Returns
    -------
    Callable
        Decorated function `f`.

    Examples
    --------
    Basic usage
        ```
        @timed
        def foo():
            ...
        foo() # prints 'foo: 5.12 ms'
        ```

    Change default name
        ```
        @timed(name='bar elapsed ')
        def bar():
            ...
        bar() # prints 'bar elapsed 2.56 ms'
        ```

    Change default units
        ```
        @timed(unit='ns')
        def baz():
            ...
        baz() # prints 'baz: 3.28e+04 ns'
        ```

    Suppress printing and use the attribute `interval` to get the time
    elapsed
        ```
        @timed(print_fn=None)
        def qux():
            ...
        qux() # prints nothing
        print(qux.interval)
        ```

    """

    def decorator(_f):
        @wraps(_f)
        def wrapped(*args, **kwargs):
            start = counter()
            return_value = _f(*args, **kwargs)
            interval = counter() - start
            wrapped.interval = interval

            if print_fn is not None:
                nonlocal name
                if name is None:
                    name = _f.__name__ + ': '
                t, u = rescale_time(interval, unit=unit)
                print_str = f'{name}{t:.3g} {u}'
                print_fn(print_str)

            return return_value

        return wrapped

    if f is None:
        return decorator
    else:
        return decorator(f)



from collections import namedtuple
from math import inf

Unit = namedtuple('Unit', 'names scale limit')

UNITS = [
    Unit(('ns',), 10 ** -9, 10 ** -6),
    Unit(('us',), 10 ** -6, 10 ** -3),
    Unit(('ms',), 10 ** -3, 1),
    Unit(('s',), 1, 10 ** 3),
    Unit(('min',), 60, 6 * 10 ** 4),
    Unit(('h', 'hs', 'hours'), 3600, 36 * 10 ** 5),
    Unit(('d', 'day', 'days'), 3600 * 24, inf)]


def rescale_time(interval: float, unit: str):
    """Rescale the time interval using the provided unit

    Parameters
    ----------
    interval: float
        Time interval to be rescaled, in seconds.
    unit: {'auto', 'ns', 'us', 'ms', 's', 'min', 'h', 'd'}
        Time unit to which `interval` should be rescaled. Use 'a' or
        'auto' for automatic time adjustment.

    Returns
    -------
    float
        Time interval in new units.
    str
        Convenient unit found automatically if input unit was 'auto'.
        Otherwise provided `unit` is returned unchanged.

    Examples
    --------
    >>> rescale_time(0.421, 'us')
    (421000.0, 'us')
    >>> rescale_time(150, 'min')
    (2.5, 'min')
    >>> rescale_time(0.911, 'auto')
    (911.0, 'ms')

    Raises
    ------
    ValueError
        If the unit provided is unknown.

    """
    unit = unit.lower().strip()

    for u in UNITS:
        if unit in u.names or (unit in ('auto', 'a') and interval < u.limit):
            return interval / u.scale, u.names[0]

    raise ValueError(f"Unknown unit: {unit}. Use one of the following: "
                     f"{[x.names[0] for x in UNITS]} or 'auto'")



from time import perf_counter as counter

from horology.tformatter import rescale_time


class Timing:
    """Context manager that measures time elapsed with the context

    Use `interval` property to get the time elapsed.

    Parameters
    ----------
    name: str, optional
        Message that should be printed before the time value, e.g.:
        'Doing x: '
    unit: str, optional
        Time unit used to print elapsed time. Possible values are:
         ['ns', 'us', 'ms', 's', 'min', 'h', 'd']. Use 'a' or 'auto'
         for automatic time adjustment (default).
    print_fn: Callable or None, optional
        Function that is called to print the time elapsed in the
        context. Use `None` to disable printing anything. You can
        provide e.g. `logger.info`. By default the built-in `print`
        function is used.

    Example
    -------
    Basic usage
        ```
        from horology import Timing
        with Timing(name='Important calculations: '):
            do_a_lot()
        ```
        Possible result:
        ```
        Important calculations: 12.4 s
        ```
    """

    def __init__(self, name=None, *, unit='a', print_fn=print):
        self.name = name if name else ""
        self.unit = unit
        self._print_fn = print_fn if print_fn else lambda _: None

        self._start = None
        self._interval = None

    @property
    def interval(self) -> float:
        """Time elapsed in seconds

        If still in the context, returns time elapsed from the moment
        of entering to the context. If the context has been already
        left, returns the total time spent in the context.

        """
        if self._interval:  # when the context exited
            return self._interval
        else:  # when still in the context
            return counter() - self._start

    def __enter__(self):
        self._start = counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._interval = self.interval
        t, u = rescale_time(self.interval, self.unit)
        print_str = f"{self.name}{t:.3g} {u}"
        self._print_fn(print_str)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['horology']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'horology',
    'version': '1.2.0',
    'description': 'Conveniently measures the time of loops, contexts and functions.',
    'long_description': '# `Horology`\n\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/mjmikulski/horology.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/mjmikulski/horology/context:python)\n[![Total alerts](https://img.shields.io/lgtm/alerts/g/mjmikulski/horology.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/mjmikulski/horology/alerts/)\n[![Downloads](https://pepy.tech/badge/horology/month)](https://pepy.tech/project/horology)\n[![PyPI version](https://badge.fury.io/py/horology.svg)](https://badge.fury.io/py/horology)\n\n[![CircleCI](https://circleci.com/gh/mjmikulski/horology/tree/master.svg?style=svg)](https://circleci.com/gh/mjmikulski/horology/tree/master)\n\nConveniently measures the time of your loops, contexts and functions.\n\n![](hourglass.jpg "Photo by Mike from Pexels")\n\n\n\n## Installation\nSimply:\n```\npip install horology\n```\n\nWorks with python versions 3.6, 3.7, 3.8 and 3.9. Tested on Linux, Windows and MacOS.\n\n## Usage\nThe following 3 tools will let you measure practically any part of your Python code.\n\n### Timing an iterable (list, tuple, generator, etc)\n#### Quick example\n```python\nfrom horology import Timed\n\nanimals = [\'cat\', \'dog\', \'crocodile\']\n\nfor x in Timed(animals):\n    feed(x)\n```\nResult:\n```\niteration    1: 12.0 s\niteration    2: 8.00 s\niteration    3: 100 s\n\ntotal 3 iterations in 120 s\nmin/median/max: 8.00/12.0/100 s\naverage (std): 40.0 (52.0) s\n\n```\n\n#### More cool stuff:\nYou can specify where (if at all) you want each iteration and summary to be printed, eg.:\n```python\nfor x in Timed(animals, unit=\'ms\', \n               iteration_print_fn=logger.debug, \n               summary_print_fn=logger.info):\n    feed(x)\n```\n\n\n### Timing a function with a `@timed` decorator\n#### Quick example\n```python\nfrom horology import timed\n\n@timed\ndef foo():\n    ...\n```\nResult:\n```\n>>> foo()\nfoo: 7.12 ms\n```\n\n#### More cool stuff:\nPersonalize time unit and name\n```python\n@timed(unit=\'s\', name=\'Processing took \')\ndef bar():\n    ...\n```\nResult:\n```\n>>> bar()\nProcessing took 0.185 s\n```\n\n\n### Timing part of code with a `Timing` context\n#### Quick example\nJust wrap your code using a `with` statement\n```python\nfrom horology import Timing\n\nwith Timing(name=\'Important calculations: \'):\n    ...\n```\nResult:\n```\nImportant calculations: 12.4 s\n```\n\n#### More cool stuff:\nYou can suppress default printing and directly use measured time (also within context)\n```python\nwith Timing(print_fn=None) as t:\n    ...\n    \nmake_use_of(t.interval)\n```\n\n\n## Time units\nTime units are by default automatically adjusted, for example you will see\n`foo: 7.12 ms` rather than `foo: 0.007 s`. If you don\'t like it, you can \noverride this by setting the `unit` argument with one of these names: \n`[\'ns\', \'us\', \'ms\', \'s\', \'min\', \'h\', \'d\']`.\n\n\n\n## Contributions \nContributions are welcomed, see [contribution guide](.github/contributing.md).\n\n\n\n## Internals\nHorology internally measures time with `perf_counter` which provides the *highest available resolution,*\n see [docs](https://docs.python.org/3/library/time.html#time.perf_counter).\n',
    'author': 'mjmikulski',
    'author_email': 'maciej.mikulski.jr@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mjmikulski/horology',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
